/**
 * Legato is a configurable, lightweight web mapping client that can be
 * easily embedded into web pages and portals, CMS and individual web
 * applications. Legato is implemented in JavaScript and based on the
 * popular open source library OpenLayers.
 *
 * Copyright (C) 2010  disy Informationssysteme GmbH, http://www.disy.net
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Copyright (c) 2010 by disy Informationssysteme GmbH
package net.disy.ogc.wps.v_1_0_0.proxy;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.MessageFormat;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import net.disy.commons.httpclient.methods.MarshallingRequestEntity;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthScope;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

// NOT_PUBLISHED
public class GenericMarshallingHttpClient {

  private static final String FILE_PROTOCOL = "file:"; //$NON-NLS-1$
  private final HttpClient httpClient;
  private HttpClient defaultHttpProxyClient;

  private final JAXBContext jaxbContext;

  private final Log log = LogFactory.getLog(getClass());
  private Map<String, ProxyConfig> multiHostProxyConfig;

  public GenericMarshallingHttpClient(JAXBContext jaxbContext) {
    this.httpClient = getHttpClientByConfig(null);
    this.jaxbContext = jaxbContext;
  }

  public Object executeFullyParametrizedGetRequest(String url)
      throws WPSConnectionException,
      WPSMarshallingException {
    log.debug(MessageFormat.format("Executing GET request: [{0}]", url));
    if (url.startsWith(FILE_PROTOCOL)) {
      log.warn(MessageFormat.format("[{0}] is a file URL, we are not using httpclient", url));
      Source input;
      try {
        input = new StreamSource(new File(new URI(url)));
        return unmarshal(input);
      }
      catch (URISyntaxException e) {
        throw new WPSConnectionException(e);
      }
      catch (JAXBException e) {
        throw new WPSMarshallingException(e);
      }
    }
    GetMethod method;
    method = new GetMethod(url);
    method.getParams().setSoTimeout(30000);
    method.setFollowRedirects(true);
    return executeAnUnmarshal(method);
  }

  private Object executeAnUnmarshal(HttpMethod method)
      throws WPSConnectionException,
      WPSMarshallingException {
    org.apache.commons.httpclient.URI uri = null;

    try {
      uri = method.getURI();
      int statusCode;
      statusCode = getHttpClientByHost(uri.getHost()).executeMethod(method);
      if (statusCode != -1) {
        InputStream responseStream = method.getResponseBodyAsStream();
        Source source = new StreamSource(responseStream);
        Object result = unmarshal(source);
        if (log.isDebugEnabled()) {
          log.debug("Received XML:"); //$NON-NLS-1$
          logJaxb(result);
        }
        return result;
      }
      return null;
    }
    catch (HttpException e) {
      throw new WPSConnectionException(MessageFormat.format(
          "Internal HttpException by the URI [{0}]: {1}",
          uri == null ? StringUtils.EMPTY : uri.toString(),
          e.getMessage()), e);
    }
    catch (IOException e) {
      throw new WPSConnectionException(MessageFormat.format(
          "Internal IOException by the URI [{0}]: {1}",
          uri == null ? StringUtils.EMPTY : uri.toString(),
          e.getMessage()), e);
    }
    catch (JAXBException e) {
      throw new WPSConnectionException(MessageFormat.format(
          "Internal JAXBException by the URI [{0}]: {1}",
          uri == null ? StringUtils.EMPTY : uri.toString(),
          e.getMessage()), e);
    }
    finally {
      method.releaseConnection();
    }
  }

  private Object unmarshal(Source source) throws JAXBException {
    Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
    Object result = unmarshaller.unmarshal(source);
    return result;
  }

  private void logJaxb(Object objToLog) throws JAXBException {
    final Marshaller marshaller = jaxbContext.createMarshaller();
    marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
    StringWriter result = new StringWriter();
    marshaller.marshal(objToLog, result);
    log.debug(result.getBuffer().toString());
  }

  public Object executePostRequest(String url, Object input) throws Exception {
    log.debug("Executing POST request on URL: [" + url + "] and with input " + input); //$NON-NLS-1$ //$NON-NLS-2$

    PostMethod method;

    try {
      Marshaller marshaller = jaxbContext.createMarshaller();

      method = new PostMethod(url);

      method.setFollowRedirects(false);

      final RequestEntity entity = new MarshallingRequestEntity(marshaller, input);
      method.setRequestEntity(entity);

      if (log.isDebugEnabled()) {
        log.debug("Sending XML"); //$NON-NLS-1$
        logJaxb(input);
      }
    }
    catch (JAXBException e) {
      throw new Exception("Internal JAXB Exception: " + e.getMessage(), e); //$NON-NLS-1$
    }

    return executeAnUnmarshal(method);

  }

  public void setDefaultProxyConfig(ProxyConfig defaultProxyConfig) {
    if (defaultProxyConfig != null) {
      this.defaultHttpProxyClient = getHttpClientByConfig(defaultProxyConfig);
    }
  }

  public void setMultiHostProxyConfig(Map<String, ProxyConfig> multiProxyHostConfig) {
    this.multiHostProxyConfig = multiProxyHostConfig;
  }

  private HttpClient getHttpClientByConfig(ProxyConfig proxyConfig) {
    HttpClient client = new HttpClient(new MultiThreadedHttpConnectionManager());
    client.getHttpConnectionManager().getParams().setConnectionTimeout(30000);

    if (proxyConfig == null) {
      return client;
    }

    client.getHostConfiguration().setProxy(proxyConfig.getHost(), proxyConfig.getPort());

    if (!StringUtils.isEmpty(proxyConfig.getUserName())) {
      client.getState().setProxyCredentials(
          new AuthScope(proxyConfig.getHost(), proxyConfig.getPort()),
          new UsernamePasswordCredentials(proxyConfig.getUserName(), proxyConfig.getPassword()));
    }

    return client;
  }

  private HttpClient getHttpClientByHost(String host) {
    if (multiHostProxyConfig != null && multiHostProxyConfig.containsKey(host)) {
      return getHttpClientByConfig(this.multiHostProxyConfig.get(host));
    }

    if (defaultHttpProxyClient != null) {
      return defaultHttpProxyClient;
    }

    return httpClient;
  }

}
